import time
from contextlib import suppress
from enum import Enum
from types import TracebackType
from typing import Optional, Type

from rich.console import Console, RenderableType
from rich.panel import Panel
from rich.status import Status


class State(Enum):
    START = 1
    STOP = 0


class ObservableStatus(Status):
    @property
    def console(self) -> "MyConsole":
        return self._live.console

    def __enter__(self) -> "Status":
        self.console.notify(self, state=State.START)
        return super().__enter__()

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        super().__exit__(exc_type, exc_val, exc_tb)
        self.console.notify(self, state=State.STOP)


class MyConsole(Console):
    """
    Extended Console with support for multiple Live components and print methods
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._status = []

    def notify(self, status_obj, state: State):
        if state == State.START:
            with suppress(IndexError):
                self._status[-1].stop()
            self._status.append(status_obj)
        else:
            self._status.remove(status_obj)
            with suppress(IndexError):
                self._status[-1].start()

    def status(
        self,
        status: RenderableType,
        *,
        spinner: str = "dots",
        spinner_style: str = "status.spinner",
        speed: float = 1.0,
        refresh_per_second: float = 12.5,
    ) -> ObservableStatus:
        """Display a status and spinner.

        Args:
            status (RenderableType): A status renderable (str or Text typically).
            spinner (str, optional): Name of spinner animation (see python -m rich.spinner). Defaults to "dots".
            spinner_style (StyleType, optional): Style of spinner. Defaults to "status.spinner".
            speed (float, optional): Speed factor for spinner animation. Defaults to 1.0.
            refresh_per_second (float, optional): Number of refreshes per second. Defaults to 12.5.

        Returns:
            Status: A Status object that may be used as a context manager.
        """

        return ObservableStatus(
            status,
            console=self,
            spinner=spinner,
            spinner_style=spinner_style,
            speed=speed,
            refresh_per_second=refresh_per_second,
        )

    def info(self, message) -> None:
        self.print(
            Panel(
                message,
                border_style="bright_blue",
                title_align="left",
                expand=True,
                title="Info",
            )
        )

    def warning(self, message) -> None:
        self.print(
            Panel(
                message,
                border_style="yellow",
                title_align="left",
                expand=True,
                title="Warning",
            )
        )


console = MyConsole()


if __name__ == "__main__":  # pragma: no cover

    with (
        console.status(
            "Preparing summary...",
        )
    ) as status:
        time.sleep(1)
        with (
            console.status(
                "Step 1",
            )
        ) as status:
            time.sleep(1)
            with (
                console.status(
                    "Step 1b",
                )
            ) as status:
                time.sleep(1)
                pass
            time.sleep(1)
        time.sleep(1)



import logging
from contextlib import suppress
from datetime import timezone, datetime
from enum import Enum
from itertools import chain
from pathlib import Path

import rich_click as click
from pydantic import BaseModel
from rich.prompt import Confirm

from .console.default import console
from .console.output import print_table_summary, print_duplicate_files
from .console.tree import DirectoryTree
from .filesystem.directory import Catalogue, Directory
from .filesystem.file import File
from .filesystem.utils import generate_filename
from .settings import GlobalSettings
from .storage import Storage

click.rich_click.SHOW_ARGUMENTS = True
# click.rich_click.GROUP_ARGUMENTS_OPTIONS = True


class Context(BaseModel):
    global_settings: GlobalSettings
    storage: Storage
    workdir: Path
    verbose: bool
    interactive: bool


class Operation(Enum):
    MOVE = "move"
    COPY = "copy"
    DELETE = "delete"

    def __str__(self):
        return self.value


@click.group(context_settings={"auto_envvar_prefix": "CATALOGUER"})
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    help="Enables verbose mode. Disabled by default",
    default=False,
)
@click.option(
    "--interactive/--no-interactive",
    help="Disables confirmation prompts. Enabled by default",
    default=True,
)
@click.pass_context
def cli(ctx, verbose, interactive):
    """
    Command line interface.

    All [OPTIONS] can be passed as environment variables with the "CATALOGUER_" prefix.

    file arguments accept file names and a special value "-" to indicate stdin or stdout
    """
    if not ctx.obj:
        global_settings = GlobalSettings()
        ctx.obj = Context(
            global_settings=global_settings,
            storage=Storage(path=global_settings.storage_location),
            workdir=Path.cwd(),
            verbose=verbose,
            interactive=interactive,
        )
        if verbose:
            console.print(ctx.obj)


@cli.command()
@click.argument("src")
@click.option(
    "--media-only/--all", help="Filter by media files. Enabled by default", default=True
)
@click.pass_obj
def inspect(ctx: Context, src, media_only):
    """
    Inspects a path or a catalogue
    """
    # TODO: allow single file
    directory = ctx.storage.load_catalogue(src, force_reload=True)
    if not directory:
        src_path = None
        with suppress(FileNotFoundError):
            src_path = Path(src).expanduser().resolve(strict=True)
        if not src_path or not src_path.is_dir():
            raise click.BadParameter(
                f'Error "{src}" is neither a catalogue or an existing directory'
            )
        directory = Directory.from_path(src_path)

    duplicated_files = directory.detect_duplicates(media_only=media_only)
    with console.status(
        "[green]Preparing summary...",
    ):
        name = directory.path
        if isinstance(directory, Catalogue):
            name = f"{directory.name} : {directory.path}"

        files = directory.files
        if media_only:
            files = [file for file in files if file.is_media_type()]
        print_table_summary(name=name, files=files, duplicated_files=duplicated_files)

        if duplicated_files:
            print_duplicate_files(
                duplicated_files=duplicated_files, from_path=directory.path
            )

    if isinstance(directory, Catalogue):
        ctx.storage.save_catalogue(directory)


@cli.command()
@click.argument("name")
@click.pass_obj
def delete_catalogue(ctx: Context, name):
    """
    Deletes a catalogue. No files are affected.
    """
    existing_catalogue = ctx.storage.load_catalogue(name)

    if not existing_catalogue:
        raise click.BadParameter(f'Catalogue "{name}" not found')

    console.warning(
        f'Are you sure to delete the catalogue "{name}" pointing to "{existing_catalogue.path}"?'
        f"\nNote: No actual files will be affected."
    )
    if ctx.interactive and not Confirm.ask(f"Do you want to proceed?"):
        return

    ctx.storage.delete_catalogue(name)


@cli.command()
@click.argument("name")
@click.argument("src", required=False)
@click.pass_obj
def create_catalogue(ctx: Context, name, src):
    """
    Creates a new catalogue.
    """
    src_path = None
    if src:
        with suppress(FileNotFoundError):
            src_path = Path(src).expanduser().resolve(strict=True)
        if not src_path or not src_path.is_dir():
            raise click.BadParameter(f'Error "{src}" is not an existing path')

    catalogue_path = src_path or ctx.workdir
    existing_catalogue = ctx.storage.load_catalogue(name)

    if existing_catalogue:
        raise click.BadParameter(
            f'There is already a catalogue named "{name}" pointing to "{existing_catalogue.path}"'
        )

    new_catalogue = Catalogue(
        name=name,
        format_pattern=ctx.global_settings.format_pattern,
        unknown_format_pattern=ctx.global_settings.unknown_format_pattern,
        path=catalogue_path,
    )
    new_catalogue.explore()

    files = new_catalogue.files
    if files:
        duplicated_files = new_catalogue.detect_duplicates()
        with console.status(
            "[green]Preparing summary...",
        ):
            print_table_summary(
                name=new_catalogue.name, files=files, duplicated_files=duplicated_files
            )

    console.print(
        f"Catalogue name: [bold purple]{name}[/]\n"
        f"Catalogue location: [bold purple]{catalogue_path}[/]\n"
        f"format_pattern: [bold purple]{new_catalogue.format_pattern}[/]\n"
        f"unknown_format_pattern: [bold purple]{new_catalogue.unknown_format_pattern}[/]\n"
    )
    ctx.storage.save_catalogue(new_catalogue)


@cli.command()
@click.argument("src")
@click.argument("dst")
@click.option("--dry-run", is_flag=True)
@click.pass_obj
def copy(ctx: Context, src, dst, dry_run):
    """
    Copy files. In case of duplicates will take the shortest name.
    """
    operation_mode = Operation.COPY
    return operate(ctx, src, dst, operation_mode, dry_run)


@cli.command()
@click.argument("src")
@click.argument("dst")
@click.option("--dry-run", is_flag=True)
@click.pass_obj
def move(ctx: Context, src, dst, dry_run):
    """
    Move files. In case of duplicates will take the shortest name.
    """
    operation_mode = Operation.MOVE
    return operate(ctx, src, dst, operation_mode, dry_run)


@cli.command()
@click.argument("src")
@click.argument("dst", required=False)
@click.option("--dry-run", is_flag=True)
@click.pass_obj
def delete_duplicates(ctx: Context, src, dst, dry_run):
    """
    Delete duplicates.
    """
    operation_mode = Operation.DELETE
    return operate(ctx, src, dst, operation_mode, dry_run)


def filter_list_of_duplicated_files(
    duplicated_list_of_files_sorted_by_name_length, files_to_process
):
    def has_different_names(files):
        return len(set([file.path.name for file in files])) > 1

    def at_least_one_file_is_going_to_be_process(files):
        return bool([file for file in files if file in files_to_process])

    return list(
        filter(
            lambda files: has_different_names(files)
            and at_least_one_file_is_going_to_be_process(files),
            duplicated_list_of_files_sorted_by_name_length,
        )
    )


def extract_files(src_data):
    duplicated_discarded_files = set()
    duplicated_list_of_files_sorted_by_name_length = []

    if isinstance(src_data, File):
        files_to_operate = [src_data]
    else:  # elif isinstance(src_data, (Catalogue, Directory)):
        duplicated_list_of_files = src_data.detect_duplicates()

        if duplicated_list_of_files:
            duplicated_list_of_files_sorted_by_name_length = list(
                sorted(duplicated_list, key=lambda file: len(file.path.name))
                for duplicated_list in duplicated_list_of_files
            )

            # Remove each first file from the duplicated list
            duplicated_discarded_files = set(
                chain(
                    *[
                        files[1:]
                        for files in duplicated_list_of_files_sorted_by_name_length
                    ]
                )
            )
        # ignore the duplicated_discarded_files since processing just one of them is enough
        files_to_operate = [
            file
            for file in src_data.files
            if file.is_media_type() and file not in duplicated_discarded_files
        ]
    return (
        duplicated_list_of_files_sorted_by_name_length,
        duplicated_discarded_files,
        files_to_operate,
    )


def get_from_input(ctx, value):
    """
    Catalogue > Directory > File > None
    """
    if not value:
        return None

    catalogue = ctx.storage.load_catalogue(value)
    if catalogue:
        return catalogue

    try:
        path = Path(value).expanduser().resolve(strict=True)
    except FileNotFoundError:
        raise click.BadParameter(
            f'Error "{value}" is neither a catalogue or an existing path'
        )

    if path.is_dir():
        return Directory.from_path(path)
    return File(path)


def operate(ctx, src, dst, operation_mode, dry_run=False):
    start_dt = datetime.now(timezone.utc)

    src_data = get_from_input(ctx, src)
    if not src_data:
        raise click.BadParameter(
            f'Error "{src}" is neither a catalogue or a valid path'
        )

    dst_data = get_from_input(ctx, dst)
    if dst_data and isinstance(dst_data, File):
        raise click.BadParameter(
            f'Error "{dst}" is neither a catalogue or an existing directory'
        )

    if isinstance(src_data, File) and (
        not dst_data or not isinstance(dst_data, (Catalogue, Directory))
    ):
        raise click.BadParameter(
            f'Error "{src}" is a file but no valid destination was provided'
        )

    with console.status(
        f"[green]Inspecting files...",
    ) as status:
        (
            duplicated_list_of_files_sorted_by_name_length,
            duplicated_discarded_files,
            files_to_operate,
        ) = extract_files(src_data)

    if operation_mode == Operation.DELETE:
        if dst_data:
            duplicate_files_across_directories = [
                file
                for file_list in dst_data.detect_duplicates_with(files_to_operate)
                for file in file_list
            ]
            # delete only dst files (which are reported as duplicated)
            files_to_process = list(
                filter(
                    lambda file: file.path.is_relative_to(dst_data.path),
                    duplicate_files_across_directories,
                )
            )
        else:  # files to delete are src duplicates
            files_to_process = [
                file for file in duplicated_discarded_files if file.is_media_type()
            ]

            duplicated_list_of_different_filenames = filter_list_of_duplicated_files(
                duplicated_list_of_files_sorted_by_name_length,
                files_to_process=files_to_process,
            )
            if duplicated_list_of_different_filenames:
                console.warning(
                    "I will keep the shortest name for each of the following groups:"
                )
                print_duplicate_files(
                    duplicated_files=duplicated_list_of_different_filenames
                )

    else:  # dst_data can only be Directory or Catalogue
        duplicate_files_across_directories = [
            file
            for file_list in dst_data.detect_duplicates_with(files_to_operate)
            for file in file_list
        ]

        files_to_process = set(files_to_operate) - set(
            duplicate_files_across_directories
        )
        duplicated_list_of_different_filenames_to_import = (
            filter_list_of_duplicated_files(
                duplicated_list_of_files_sorted_by_name_length,
                files_to_process=files_to_process,
            )
        )
        if duplicated_list_of_different_filenames_to_import:
            console.warning(
                "The files you attempt to import contain duplicates with different names.\n"
                "I will use the shortest name for each of the following groups:"
            )
            print_duplicate_files(
                duplicated_files=duplicated_list_of_different_filenames_to_import
            )

    console.info(f"Detected {len(files_to_process)} files.")
    if dry_run:
        console.warning(f"Running in dry-run, so no changes will be effective.")
    if (
        files_to_process
        and ctx.interactive
        and not Confirm.ask(f"Do you want to proceed to {operation_mode} these files?")
    ):
        return

    # actual processing
    tree, skipped_tree = process_files(
        ctx, src_data, dst_data, files_to_process, operation_mode, start_dt, dry_run
    )

    if dst_data:
        tree_starting_path = dst_data.path
    else:
        tree_starting_path = src_data.path

    guide_style = "bold bright_blue"
    if operation_mode == Operation.DELETE:
        guide_style = "bold green"
    rendered_tree = tree.generate_tree(tree_starting_path, guide_style=guide_style)
    if rendered_tree:
        console.print(f"\n{operation_mode.value.title()} {tree.file_count} files:")
        console.print(rendered_tree)

    rendered_skipped_tree = skipped_tree.generate_tree(
        src_data.path, guide_style="bold bright_black"
    )
    if rendered_skipped_tree:
        console.print(
            f"\n:warning: {skipped_tree.file_count} files have not being processed:"
        )
        console.print(rendered_skipped_tree)

    # save catalogues
    if isinstance(src_data, Catalogue) and not dry_run:
        ctx.storage.save_catalogue(src_data)
    if isinstance(dst_data, Catalogue) and not dry_run:
        ctx.storage.save_catalogue(dst_data)


def process_files(
    ctx, src_data, dst_data, files_to_process, operation_mode, start_dt, dry_run
):
    path_format = ctx.global_settings.format_pattern
    unknown_format_pattern = ctx.global_settings.unknown_format_pattern
    if isinstance(dst_data, Catalogue):
        path_format = dst_data.format_pattern
        unknown_format_pattern = dst_data.unknown_format_pattern

    tree = DirectoryTree()
    skipped_tree = DirectoryTree()

    with console.status(
        f"[green]Processing files...",
    ) as status:
        for file in files_to_process:
            status.update(
                status=f"[green]Processing file {tree.file_count} of {len(files_to_process)}"
            )
            dst_file_path = None
            old_path = file.path

            if operation_mode != Operation.DELETE:
                new_filename = generate_filename(
                    file,
                    src_data,
                    unknown_format_pattern=unknown_format_pattern,
                    path_format=path_format,
                    import_dt=start_dt,
                )
                if not new_filename:
                    skipped_tree.add_imported_file(file, old_path=file.path)
                    continue

                dst_file_path = dst_data.path.joinpath(new_filename)

            processed_file = process_file(
                file,
                dst_file_path,
                operation=operation_mode,
                dst_directory=dst_data,
                dry_run=dry_run,
            )

            tree.add_imported_file(processed_file, old_path=old_path)
    return tree, skipped_tree


def process_file(file, dst_file_path, operation, dst_directory, dry_run):
    logging.debug(f"{file.path} -> {dst_file_path}")

    if operation == Operation.DELETE:
        if not dry_run:
            file.delete()
        return file

    path_available = dst_directory.is_path_available(dst_file_path)
    if not path_available:
        logging.debug(f"Path {dst_file_path} not available, renaming file")
        dst_file_path = dst_directory.find_new_path(dst_file_path)

    if dry_run:
        file._path = dst_file_path
        return file

    if not dst_file_path.parent.exists():
        logging.debug(f'Creating folder "{dst_file_path.parent}"')
        dst_file_path.parent.mkdir(parents=True)

    if operation == Operation.MOVE:
        file.move_file(dst_file_path)
        dst_directory.add_file(file)
        return file

    if operation == Operation.COPY:
        new_file = file.clone_file(dst_file_path)
        dst_directory.add_file(new_file)
        return new_file


if __name__ == "__main__":
    cli()



from pathlib import Path
from typing import List, Set

from pydantic import BaseModel
from rich.markup import escape
from rich.tree import Tree

from ..filesystem.file import File
from ..filesystem.utils import approximate_size


class FileInfo(BaseModel):
    file: File
    old_path: Path

    def __hash__(self):
        return self.file.path.__hash__()

    def __eq__(self, other):
        return self.file.path == other.file.path

    class Config:
        arbitrary_types_allowed = True


class DirectoryInfo(BaseModel):
    path: Path
    files: Set[FileInfo]
    sub_directories: Set["DirectoryInfo"]

    def __hash__(self):
        return self.path.__hash__()

    def __eq__(self, other):
        return self.path == other.path


class DirectoryTree:
    tree: dict
    file_count: int = 0

    def __init__(self):
        self.tree = {}

    def add_imported_file(self, file, old_path):
        parent_directory_info = None

        if file.path is None:
            path = old_path  # file was deleted
        else:
            path = file.path

        for directory_path in _get_parents(path.parent):
            directory_info = self.tree.setdefault(
                directory_path,
                DirectoryInfo(path=directory_path, files=[], sub_directories=[]),
            )
            if parent_directory_info:
                parent_directory_info.sub_directories.add(directory_info)
            parent_directory_info = directory_info

        self.file_count += 1
        directory_info.files.add(
            FileInfo(
                file=file,
                old_path=old_path,
            )
        )

    def generate_tree(self, dst_path, guide_style="bold bright_blue"):
        directory_info = self.tree.get(dst_path)
        if not directory_info:
            return None
        tree = Tree(
            f":open_file_folder: [link file://{directory_info.path}]{escape(directory_info.path.name)}",
            guide_style=guide_style,
        )
        _expand_tree(tree=tree, directory_info=directory_info)
        return tree


def _expand_tree(tree, directory_info: DirectoryInfo):
    for file_info in directory_info.files:
        if file_info.file.path is None:
            tree.add(
                f"{file_info.old_path.name} [b green]{approximate_size(file_info.file.size)}"
            )
        elif file_info.old_path.name != file_info.file.path.name:
            tree.add(
                f"{file_info.file.path.name} ({file_info.old_path.name}) [b bright_yellow]{approximate_size(file_info.file.size)}"
            )
        else:
            tree.add(
                f"{file_info.file.path.name} [b bright_yellow]{approximate_size(file_info.file.size)}"
            )

    for sub_directory_info in sorted(
        directory_info.sub_directories, key=lambda directory: directory.path
    ):
        branch = tree.add(
            f":open_file_folder: [b bold][link file://{sub_directory_info.path}]{escape(sub_directory_info.path.name)}[/]"
        )
        _expand_tree(tree=branch, directory_info=sub_directory_info)


def _get_parents(path: Path) -> List[Path]:
    directories = []
    while path != Path("/"):
        directories.append(path)
        path = path.parent
    directories.reverse()
    return directories



import json
import logging
import os
from contextlib import suppress

from datetime import datetime, timezone
from itertools import chain
from pathlib import Path
from typing import List, Dict, Optional
from rich.progress import (
    Progress,
    TextColumn,
    BarColumn,
    TaskProgressColumn,
    MofNCompleteColumn,
    TimeRemainingColumn,
    SpinnerColumn,
    TimeElapsedColumn,
)

from .file import File
from .utils import split_extension_from_filename, count_number_of_files
from ..console.default import console

DATABASE_LOCATION = ".cataloguer_db.json"

logger = logging.getLogger(__name__)


class Directory:
    path = None
    _files: List[File] = None
    _files_by_path: Dict[Path, File] = None
    _files_by_size: Dict[int, File] = None

    @property
    def files(self):
        return tuple(self._files.copy())

    @files.setter
    def files(self, value):
        list(map(self.add_file, value))
        self._files = value

    def __init__(self, path: Path, files: Optional[List[File]] = None):
        self.path = path.resolve()
        self._files_by_path = {}
        self._files_by_size = {}
        self.files = files or []

    @classmethod
    def from_path(cls, path: Path):
        directory = cls(path=path)
        directory.explore()
        return directory

    def explore(self):
        files = []

        # with Progress(
        #     TextColumn(f"[green]Exploring {self.path.name}..."),
        #     BarColumn(),
        #     TextColumn("[progress.description]{task.description}"),
        # ) as progress:
        #     discovery_task = progress.add_task("Exploring repo", total=None)

        with console.status(
            f"[green]Exploring {self.path.name}...",
        ) as status:
            for dirpath, dirnames, filenames in os.walk(self.path):
                for filename in filenames:
                    full_path = os.path.join(dirpath, filename)
                    try:
                        # if the target is a symlink (soft one), this will
                        # dereference it - change the value to the actual target file
                        file_path = Path(os.path.realpath(full_path))
                        file_size = os.path.getsize(file_path)
                    except OSError as e:
                        # not accessible (permissions, etc) - pass on
                        logger.warning("Cannot read %s: %s", full_path, e)
                        continue
                    files.append(File(path=file_path, size=file_size))
                    # progress.update(discovery_task, description=f"Found {len(files)} files")
                status.update(
                    status=f"[green]Exploring {self.path.name}. Found {len(files)} files"
                )
            # progress.update(discovery_task, total=len(files), completed=len(files))

        self.files = files
        return files

    def notify(self, file, field, new_value):
        """
        Observer notification method
        """
        if field == "path":
            del self._files_by_path[file.path]
            # if not new_value.is_relative_to(self.path): # New in version 3.9
            if not str(new_value or "").startswith(str(self.path)):
                file.unsubscribe(self)
                with suppress(ValueError):
                    self._files.remove(file)
                with suppress(ValueError):
                    self._files_by_size.setdefault(file.size, []).remove(file)
                    if not self._files_by_size[file.size]:
                        del self._files_by_size[file.size]
                return
            self._files_by_path[new_value] = file

    def add_file(self, file):
        file.subscribe(self)
        self._files.append(file)
        self._files_by_path[file.path] = file
        self._files_by_size.setdefault(file.size, []).append(file)

    @staticmethod
    def detect_duplicates_on_files(files_by_size) -> List[List[File]]:
        _files_by_hash = {}
        _files_by_short_hash = {}
        file_size_collisions = filter(
            lambda items: len(items) > 1, files_by_size.values()
        )
        file_size_collisions = list(chain(*file_size_collisions))

        # with Progress(
        #     TextColumn("[progress.description]{task.description}"),
        #     BarColumn(),
        #     TaskProgressColumn(),
        #     MofNCompleteColumn(),
        #     TimeRemainingColumn(),
        #     # expand=True,
        #     # transient=True
        # ) as progress:
        #     file_size_collisions = list(chain(*file_size_collisions))
        #     fast_detection_task = progress.add_task("[green]Quick file inspection", total=len(file_size_collisions))
        #     slow_detection_task = progress.add_task("[green]Inspecting full files", total=None)
        with console.status(f"[green]Inspecting files for duplication...") as status:
            for index, file in enumerate(file_size_collisions, start=1):
                # progress.update(fast_detection_task, completed=index)
                short_file_hash = file.short_hash
                _files_by_short_hash.setdefault(short_file_hash, []).append(file)

            short_file_hash_collisions = filter(
                lambda items: len(items) > 1, _files_by_short_hash.values()
            )

            short_file_hash_collisions = list(chain(*short_file_hash_collisions))
            # progress.update(slow_detection_task, total=len(short_file_hash_collisions))
            for index, file in enumerate(short_file_hash_collisions, start=1):
                # progress.update(slow_detection_task, completed=index)
                status.update(
                    status=f"[green]Inspecting file {index} of {len(short_file_hash_collisions)} for duplicates"
                )
                file_hash = file.hash
                _files_by_hash.setdefault(file_hash, []).append(file)

            file_hash_collisions = filter(
                lambda items: len(items) > 1, _files_by_hash.values()
            )
            return list(file_hash_collisions)

    def detect_duplicates(self, media_only=True):
        files_by_size = self._files_by_size
        if media_only:
            files_by_size = {
                size: files
                for size, files in self._files_by_size.items()
                if files[0].is_media_type()
            }
        return self.detect_duplicates_on_files(files_by_size=files_by_size)

    def detect_duplicates_with(self, files, media_only=True):
        def intersect(a, b):
            return {k: list(chain(a[k], b[k])) for k in a.keys() & b.keys()}

        files_by_size = self._files_by_size
        if media_only:
            files_by_size = {
                size: files
                for size, files in self._files_by_size.items()
                if files[
                    0
                ].is_media_type()  # if one of them is media type, all are since are duplicates
            }

        given_files_by_size = {}
        for file in files:
            given_files_by_size.setdefault(file.size, []).append(file)

        intersection_of_files_by_size = intersect(files_by_size, given_files_by_size)
        return self.detect_duplicates_on_files(
            files_by_size=intersection_of_files_by_size
        )

    def is_path_available(self, path):
        return self._files_by_path.get(path) is None

    def find_new_path(self, path):
        basename, filename_extension = split_extension_from_filename(path.name)
        i = 0
        while True:
            i += 1
            new_filename = f"{basename}_{i}.{filename_extension}"
            new_path = Path(path.parent.joinpath(new_filename))
            if self._files_by_path.get(new_path) is None:
                return new_path


class Catalogue(Directory):
    name: str
    creation_date: datetime
    format_pattern: str
    unknown_format_pattern: Optional[str]

    def __init__(
        self,
        name: str,
        format_pattern: str,
        unknown_format_pattern: Optional[str] = None,
        creation_date: datetime = None,
        **kwargs,
    ):
        self.name = name
        self.format_pattern = format_pattern
        self.unknown_format_pattern = unknown_format_pattern
        self.creation_date = creation_date or datetime.now(timezone.utc)
        super().__init__(**kwargs)

    def dict(self):
        def file_asdict(file):
            file_dict = file.asdict()
            file_dict["path"] = str(file.path.relative_to(self.path))
            return file_dict

        return {
            "name": self.name,
            "path": self.path.resolve(),
            "creation_date": self.creation_date.isoformat(),
            "format_pattern": self.format_pattern,
            "unknown_format_pattern": self.unknown_format_pattern,
            "files": [file_asdict(file) for file in self._files],
        }

    def save(self, path: Path):
        catalogue_data = self.dict()
        with open(path, "w") as fd:
            json.dump(catalogue_data, fd, default=str)

    @classmethod
    def parse_obj(cls, data, force_reload=False):
        path = Path(data["path"]).resolve(strict=True)
        creation_date = datetime.fromisoformat(data["creation_date"])
        catalogue = cls(
            name=data["name"],
            path=path,
            creation_date=creation_date,
            format_pattern=data["format_pattern"],
            unknown_format_pattern=data.get("unknown_format_pattern"),
        )
        files_on_path = count_number_of_files(path)
        logger.debug(
            f"Catalogue files: {len(data['files'])} vs filesystem files {files_on_path}"
        )
        if force_reload or len(data["files"]) != files_on_path:
            catalogue.explore()
        else:
            files = [
                File(**{**file_data, "path": path.joinpath(file_data["path"])})
                for file_data in data["files"]
            ]
            catalogue.files = files
        return catalogue



from collections import Counter, defaultdict
from pathlib import Path

from rich import box
from rich.columns import Columns
from rich.markdown import Markdown
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

from ..console.default import console
from ..filesystem.utils import approximate_size


def print_processing(src_path: Path, dst_path: Path, imported_files, skipped_files):
    table = Table(
        show_header=True,
        show_footer=True,
        header_style="bold",
        box=box.SIMPLE,
        title=f"Processed files",
        # caption=f"Path: [b magenta]{catalogue.path}[/]"
    )
    table.border_style = "bright_yellow"
    table.add_column("Original Path", style="white")
    table.add_column("Destination Path", style="white")

    for original_path, file in imported_files:
        table.add_row(
            str(original_path.relative_to(src_path)),
            str(file.path.relative_to(dst_path)),
        )
    console.print(table)


def print_table_summary(duplicated_files, files, name):
    files_by_media = Counter((file.get_media_type() for file in files))
    files_size_by_media = defaultdict(int)
    for file in files:
        files_size_by_media[file.get_media_type()] += file.size
    duplicated_files_by_media = Counter(
        (file.get_media_type() for sublist in duplicated_files for file in sublist[:1])
    )
    table = Table(
        show_header=True,
        show_footer=True,
        header_style="bold",
        box=box.SIMPLE,
        # title=f'Catalogue: [b blue]{catalogue.name}',
        # caption=f"Path: [b magenta]{catalogue.path}[/]"
    )
    table.border_style = "bright_black"
    table.add_column(
        "Media Type", Text.from_markup("[b]Total", justify="right"), style="white"
    )
    table.add_column(
        "Size",
        approximate_size(sum(files_size_by_media.values())),
        justify="right",
        style="white",
        no_wrap=True,
    )
    table.add_column(
        "Files", str(len(files)), justify="right", style="white", no_wrap=True
    )
    table.add_column(
        "Duplicates",
        Text.from_markup(str(sum(duplicated_files_by_media.values())), style="red"),
        no_wrap=True,
        justify="right",
        style="red",
    )
    for file_type, count in files_by_media.items():
        prefix = "[dim]"
        if file_type in ("image", "video"):
            prefix = ""
        table.add_row(
            prefix + file_type,
            prefix + approximate_size(files_size_by_media[file_type]),
            prefix + str(count),
            prefix + str(duplicated_files_by_media.get(file_type, 0)),
        )
    # centered_table = Align.center(table)
    console.print(Markdown(f"# {name}"))
    console.print(table)


def print_duplicate_files(duplicated_files, from_path=None):
    console.print(Markdown("## Duplicates"))
    panels = []
    for duplicated_list in sorted(
        duplicated_files,
        key=lambda duplicated_list: duplicated_list[0].size,
        reverse=True,
    ):
        if from_path:
            filenames = [
                str(file.path.relative_to(from_path)) for file in duplicated_list
            ]
        else:
            filenames = [file.path.name for file in duplicated_list]

        items = []
        for path, count in Counter(filenames).items():
            text = path
            if count > 1:
                text = f"{path} (x{count})"
            items.append(text)
        panels.append(
            Panel(
                "\n".join(items),
                border_style="bright_black",
                expand=True,
                title=approximate_size(duplicated_list[0].size),
            )
        )
    console.print(Columns(panels))



import json
import logging
from pathlib import Path

from pydantic import BaseModel

from .filesystem.directory import Catalogue

logger = logging.getLogger(__name__)


class Storage(BaseModel):
    path: Path

    def load_catalogue(self, name: str, force_reload=True):
        try:
            with open(self.path.joinpath(f"{name}.json"), "r") as fd:
                return Catalogue.parse_obj(json.load(fd), force_reload=force_reload)
        except FileNotFoundError:
            return None
        except Exception as exception:
            logger.warning(f'Error happen when loading "{name}": {exception}')
            return None

    def delete_catalogue(self, name: str):
        self.path.joinpath(f"{name}.json").unlink()

    def save_catalogue(self, catalogue: Catalogue):
        catalogue.save(self.path.joinpath(f"{catalogue.name}.json"))



import datetime
import logging
import re

import PIL.ExifTags
import dateutil.parser
from PIL import Image
from PIL.TiffTags import TAGS


DATE_PATH_REGEXES = (
    r"/(?P<year>\d{4})-(?P<month>\d{2})-(?P<day>\d{2})",
    r"/IMG_(?P<year>\d{4})(?P<month>\d{2})(?P<day>\d{2})_",
    r"/VID_(?P<year>\d{4})(?P<month>\d{2})(?P<day>\d{2})_",
    r"/IMG-(?P<year>\d{4})(?P<month>\d{2})(?P<day>\d{2})-",
)


def get_path_creation_date(path):
    for regex in DATE_PATH_REGEXES:
        match = re.search(regex, str(path))
        if match:
            year, month, day = match.groups()
            try:
                return datetime.datetime(year=int(year), month=int(month), day=int(day))
            except ValueError:
                pass
    return None


def get_image_creation_date(path):
    try:
        image = Image.open(path)
    except IOError as e:
        logging.warning(e)  # cannot identify image file
        return None

    try:
        metadata = _get_exif(image)
    except AttributeError as e:
        logging.debug(f"Cannot get exif metadata for {path}")
        return None

    try:
        return _extract_created_date_from_exif(metadata)
    except ValueError:
        logging.debug(f"Cannot get creation date from exif metadata for {path}")
        return None


def _get_exif(image):
    if image.format == "TIFF":
        return {TAGS.get(key): image.tag[key] for key in image.tag.keys()}
    return {
        PIL.ExifTags.TAGS[k]: v
        for k, v in image._getexif().items()
        if k in PIL.ExifTags.TAGS
    }


def _normalize_datetime_format(exif_dt_field):
    exif_dt = exif_dt_field
    if type(exif_dt_field) in (tuple, list):
        if len(exif_dt_field) == 1:
            exif_dt = exif_dt_field[0]
    try:
        # format YYYY:MM:DD hh:mm:ss is not parsed correctly by dateutil
        return datetime.datetime.strptime(exif_dt, "%Y:%m:%d %H:%M:%S").isoformat()
    except ValueError:
        return exif_dt
    except TypeError:
        logging.debug(f"Cannot parse {exif_dt=}")
        return None


def _extract_created_date_from_exif(exif):
    exif_dt_field = exif.get("DateTimeOriginal") or exif.get("DateTime")

    if not exif_dt_field:
        return None

    created_data = _normalize_datetime_format(exif_dt_field)
    if not created_data:
        return None

    try:
        return dateutil.parser.parse(created_data)
    except ValueError as e:
        if e.args[0] == "Unknown string format: %s":
            unknown_date = e.args[1]
            logging.debug(f"Attempting to parse unknown date {unknown_date}")
            match = re.search(
                r"(?P<year>\d{4})/(?P<month>\d{2})/(?P<day>\d{2})", unknown_date
            )
            if match:
                year, month, day = match.groups()
                return datetime.datetime(year=int(year), month=int(month), day=int(day))
        raise e



import hashlib
import os
from pathlib import Path


DATABASE_LOCATION = ".cataloguer_db.json"

KEEP_PARENT_FOLDERS = (  # TODO: undocumented behaviour
    "Time Lapse",
    "Burst Sequence",
)

UNITS = {1000: ["KB", "MB", "GB"], 1024: ["KiB", "MiB", "GiB"]}


def split_extension_from_filename(filename: str):
    name_split = filename.split(".")
    if len(name_split) > 1:
        return (".".join(name_split[0:-1]), name_split[-1])
    return (filename, "")


def _keep_parent_directory(path) -> str:
    parent_folder = path.parent.name
    if any(map(lambda pattern: parent_folder.startswith(pattern), KEEP_PARENT_FOLDERS)):
        return parent_folder
    return ""


def generate_filename(file, scr_data, unknown_format_pattern, path_format, import_dt):
    media_type, media_format = file.get_type()

    # FIXME: obscure logic to keep gopro directories
    # FIXME: bug if filename + relative_path is used
    parent_directory = _keep_parent_directory(file.path)
    relative_path = str(file.path.relative_to(scr_data.path).parent)

    uses_date = bool("%" in path_format)
    if not uses_date:
        return _generate_filename(
            file,
            path_format,
            dt=None,
            parent_directory=parent_directory,
            media_type=media_type,
            media_format=media_format,
            relative_path=relative_path,
        )

    created = file.get_creation_date()
    if created:
        return _generate_filename(
            file,
            path_format,
            dt=created,
            parent_directory=parent_directory,
            media_type=media_type,
            media_format=media_format,
            relative_path=relative_path,
        )

    if unknown_format_pattern:
        return _generate_filename(
            file,
            unknown_format_pattern,
            dt=import_dt,
            parent_directory=parent_directory,
            media_type=media_type,
            media_format=media_format,
            relative_path=relative_path,
        )

    return None


def _generate_filename(
    file, path_format, dt, parent_directory, media_type, media_format, relative_path
):
    file_name, file_extension = file.split_extension()
    name = Path(parent_directory or "").joinpath(file.path.name)
    strftime_format = path_format.format(
        file=name,
        file_extension=file_extension,
        file_name=file_name,
        media_type=media_type,
        media_format=media_format,
        relative_path=relative_path,
    )
    if dt:
        return dt.strftime(strftime_format)
    return strftime_format


def count_number_of_files(path):
    file_count = 0
    for dirpath, dirnames, filenames in os.walk(path):
        for _ in filenames:
            file_count += 1
    return file_count


def _chunk_reader(fobj, chunk_size=1024):
    """Generator that reads a file in chunks of bytes"""
    while True:
        chunk = fobj.read(chunk_size)
        if not chunk:
            return
        yield chunk


def get_hash(path, first_chunk_only=False):
    hash_obj = hashlib.sha1()
    with open(path, "rb") as file_object:
        if first_chunk_only:
            hash_obj.update(file_object.read(1024))
        else:
            for chunk in _chunk_reader(file_object):
                hash_obj.update(chunk)
    return hash_obj.hexdigest()


def approximate_size(size, international_system=True):
    mult = 1000 if international_system else 1024
    for unit in UNITS[mult]:
        size = size / mult
        if size < mult:
            return "{0:.2f} {1}".format(size, unit)



from pathlib import Path
from typing import Optional

import click
from pydantic import BaseSettings, validator


ALLOWED_FORMAT_VARIABLES = (
    "media_type",
    "media_format",
    "file",
    "file_extension",
    "file_name",
    "relative_path",
)


class GlobalSettings(BaseSettings):
    format_pattern: str = "%Y/%m/%d/{file}"
    unknown_format_pattern: Optional[str] = None
    storage_location: Path = Path.home().joinpath(".catalogues/")

    class Config:
        env_prefix = "CATALOGUER_"

    @validator("format_pattern", "unknown_format_pattern", allow_reuse=True)
    def validate_format_pattern(cls, format_pattern: str):
        if format_pattern:
            try:
                format_pattern.format(
                    **{field: "" for field in ALLOWED_FORMAT_VARIABLES}
                )
            except KeyError as exception:
                raise click.BadParameter(
                    f'Unrecognised "{exception.args[0]}" format variable. Please check your format pattern.'
                )
        return format_pattern

    @validator("storage_location")
    def storage_location_must_exists(cls, storage_location: Path):
        if storage_location:
            storage_location = storage_location.resolve()
            if storage_location.exists() and not storage_location.is_dir():
                raise ValueError(f"{storage_location} is not a directory")
            storage_location.mkdir(parents=True, exist_ok=True)
        return storage_location



import shutil
from contextlib import suppress
from pathlib import PurePath, Path

import magic

from .metadata import get_image_creation_date, get_path_creation_date
from .utils import get_hash, split_extension_from_filename


class Observable:
    def __init__(self):
        self._observers = set()

    def subscribe(self, observer):
        self._observers.add(observer)

    def unsubscribe(self, observer):
        with suppress(ValueError):
            self._observers.remove(observer)

    def notify(self, *args):
        for observer in list(self._observers):
            observer.notify(self, *args)


class File(Observable):
    _path: Path
    size: int
    _hash: int
    _short_hash: int

    def __init__(self, path, size=None, hash=None, short_hash=None):
        super().__init__()
        if not isinstance(path, PurePath):
            path = Path(path)
        self._path = path
        self.size = size or path.stat().st_size
        self._hash = hash
        self._short_hash = short_hash

    def __str__(self):
        return str(self.path)

    @property
    def path(self):
        return self._path

    @path.setter
    def path(self, value):
        self.notify("path", value)
        self._path = value

    @property
    def hash(self):
        if self._hash is None:
            self.hash = get_hash(self.path)
        return self._hash

    @hash.setter
    def hash(self, value):
        self.notify("hash", value)
        self._hash = value

    @property
    def short_hash(self):
        if self._short_hash is None:
            self.short_hash = get_hash(self.path, first_chunk_only=True)
        return self._short_hash

    @short_hash.setter
    def short_hash(self, value):
        self.notify("short_hash", value)
        self._short_hash = value

    def clone_file(self, new_path):
        # if new_path.exists():
        #     raise FileExistsError
        shutil.copy2(str(self.path), str(new_path))
        return File(
            path=new_path, size=self.size, hash=self._hash, short_hash=self._short_hash
        )

    def move_file(self, new_path):
        # if new_path.exists():
        #     raise FileExistsError
        shutil.move(self.path, new_path)
        self.path = new_path

    def delete(self):
        self.path.unlink()
        # TODO: delete parent folder if is empty too?
        self.path = None

    def split_extension(self):
        return split_extension_from_filename(self.path.name)

    def get_creation_date(self):
        creation_date = None
        if self.is_image():
            creation_date = get_image_creation_date(self.path)
        if not creation_date:
            creation_date = get_path_creation_date(self.path)
        return creation_date

    def is_media_type(self):
        return self.is_image() or self.is_video()

    def is_image(self):
        return self.get_media_type() == "image"

    def is_video(self):
        return self.get_media_type() == "video"

    def get_media_type(self):
        media_type, _ = self.get_type()
        return media_type

    def get_type(self):
        mimetype = magic.from_file(str(self.path), mime=True)
        media_type = mimetype.split("/")[0]
        media_format = "/".join(mimetype.split("/")[1:])
        return media_type, media_format

    def asdict(self):
        return {
            "path": str(self._path),
            "size": self.size,
            "hash": self._hash,
            "short_hash": self._short_hash,
        }



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cataloguer', 'cataloguer.console', 'cataloguer.filesystem']

package_data = \
{'': ['*']}

install_requires = \
['Pillow==8.1.2',
 'click>=8.1.3,<9.0.0',
 'pydantic>=1.10.5,<2.0.0',
 'python-dateutil==2.8.2',
 'python-magic==0.4.27',
 'rich-click>=1.6.1,<2.0.0']

entry_points = \
{'console_scripts': ['cataloguer = cataloguer.cli:cli']}

setup_kwargs = {
    'name': 'cataloguer',
    'version': '2.0.1',
    'description': 'Organize your media files',
    'long_description': '# Media cataloguer\n\nOrganize your media files using your preferred directory structure.\n\n* Do you have pictures taken from different devices? \n* Do you want control over your file system?\n* Do you wish to unify all your media files and remove duplicates?\n\nIf you reply affirmative to any of the previous questions, this tool could be right for you too.\n\nThis tool helps you to unify and organise your media files using your own rules. \nIt also deals with duplicates, so you don\'t have to.\n\n\n## Features\n\n* Move, Copy or delete duplicates operations\n* User-friendly console output\n* Obscure creation date detection\n* Custom folder structure definition\n* Duplication detection\n* Does not alter existing files\n\n\n## Requirements\n- Python 3.9 or higher\n\n### In a virtual environment\n\n    poetry install\n    cataloguer --help\n\n\n### Usage\n\n```bash\n$ cataloguer --help\n\n Usage: cataloguer [OPTIONS] COMMAND [ARGS]...                                                                                                                                                \n                                                                                                                                                                                             \n Command line interface.                                                                                                                                                                     \n All [OPTIONS] can be passed as environment variables with the "CATALOGUER_" prefix.                                                                                                          \n file arguments accept file names and a special value "-" to indicate stdin or stdout                                                                                                        \n                                                                                                                                                                                             \n╭─ Options ─────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────╮\n│ --verbose                       -v    Enables verbose mode. Disabled by default                                                                                                           │\n│ --interactive/--no-interactive        Disables confirmation prompts. Enabled by default                                                                                                   │\n│ --help                                Show this message and exit.                                                                                                                         │\n╰───────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────╯\n╭─ Commands ────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────╮\n│ copy                                       Copy files. In case of duplicates will take the shortest name.                                                                                 │\n│ create-catalogue                           Creates a new catalogue.                                                                                                                       │\n│ delete-catalogue                           Deletes a catalogue. No files are affected.                                                                                                    │\n│ delete-duplicates                          Delete duplicates.                                                                                                                             │\n│ inspect                                    Inspects a path or a catalogue                                                                                                                 │\n│ move                                       Move files. In case of duplicates will take the shortest name.                                                                                 │\n╰───────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────╯\n```\n\n\n## Quickstart\n\nWe are going to start creating a new directory `media`:\n\n    mkdir media\n\nWe are going to create a new catalogue named `local_photos` which is going to get store on the `media` directory.\nWe specify our format pattern so photos are group by `year` and a subgroup of `month`:\n\n    export CATALOGUER_FORMAT_PATTERN=%Y/%m/{file}\n    cataloguer create-catalogue local_media --path ./media \n\n\nNow, we add some photos from an old storage driver:\n\n    cataloguer copy /mnt/hdd/old_photos local_media\n\n\nLater on, we decided we want to reorganize our local home folder, but we are not sure of how many files are \ngoing to be affected, so we run the command in `dry-run` mode:\n\n    cataloguer move ~/ local_media --dry-run\n\nAfter seeing the output, we decided to just reorganize our `Pictures`:\n\n    cataloguer move ~/Pictures/ local_media\n\n\nTo get a summary of our catalogue we run:\n\n    cataloguer inspect local_media\n\n\n## Options\n\n`CATALOGUER_FORMAT_PATTERN` accepts the following patterns\n* Common date codes:\n  * `%d`: Day of the month as number\n  * `%m`: Month as number\n  * `%Y`: Year as number\n  * `%A`: Weekday name \n  * `%B`: Month name\n  * other format codes specified [here](https://docs.python.org/3/library/datetime.html#strftime-and-strptime-format-codes)\n* `/` Specifies a new folder\n* `{media_type}`: File type (`image`, `video`)\n* `{media_format}`: File format (`jpeg`, `tiff`, `png`, `gif`, `mp4` ...)\n* `{file}` Original filename\n* `{file_extension}` Original filename extension \n* `{file_name}` Original filename without the extension\n* `{relative_path}` Relative path to the source directory\n\n\n### Advance usage:\n\n`CATALOGUER_UNKNOWN_PATH_FORMAT` Accepts the same variables as `CATALOGUER_FORMAT_PATTERN` but date patterns \nare resolved using the current date since it was not possible to recover the creation date of the file.\nThis can be useful to not leave files behind.\n\n`CATALOGUER_STORAGE_LOCATION` Accepts any path. That location will store cataloguer metadata.\nBy default, it will create a `.catalogues` in the user\'s home directory.\n\n#### Examples:\n\nPattern to fix file extensions keeping the folder structure:\n\n    CATALOGUER_FORMAT_PATTERN={relative_path}/{basename}.{media_format} cataloguer ./input ./output\n\n\n## TODO list\n\n* Video support\n',
    'author': 'Iago Veloso',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/iago1460/cataloguer',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9',
}


setup(**setup_kwargs)
